/*
 * Copyright (c) Huawei Technologies Co., Ltd. 2021-2021. All rights reserved.
 * Description:MViz Image数据类型示例代码
 * Create: 2021-08-12
 */
#include "image.h"
#include "ap_utils.h"
#include <sys_ctr.h>
#include <time.h>

#define FILE_TEST 0
#define CAMERA_4 4
#define CAMERA_5 5
#define CAMERA_6 6
#define CAMERA_7 7


namespace {
ara::core::Vector<uint8_t> imageData {};

std::vector<mdc::camera::CameraAccessConfig> camera_configs = {
		{
				.moduleName = "HUAWEI_BU18TM41_OX03C10",
#ifdef CONF_SOC_A
				.cameraId = 8U,
				.fov = 200U,
#else
				.cameraId = 4U,
				.fov = 200U,
#endif
				.fps = 30,
				.width = 1920U,
				.height = 1080U,
				.streamDepth = 5U,
				.clusterId = 0U,

		}
		,{
				.moduleName = "HUAWEI_BU18TM41_OX03C10",
#ifdef CONF_SOC_A
				.cameraId = 9U,
				.fov = 200U,
#else
				.cameraId = 5U,
				.fov = 200U,
#endif
				.fps = 30,
				.width = 1920U,
				.height = 1080U,
				.streamDepth = 3U,
				.clusterId = 1U,
		},
		{
				.moduleName = "HUAWEI_BU18TM41_OX03C10",
#ifdef CONF_SOC_A
				.cameraId = 10U,
				.fov = 200U,
#else
				.cameraId = 6U,
				.fov = 200U,
#endif
				.fps = 30,
				.width = 1920U,
				.height = 1080U,
				.streamDepth = 5U,
				.clusterId = 2U,
		}
		,{
				.moduleName = "HUAWEI_BU18TM41_OX03C10",
#ifdef CONF_SOC_A
				.cameraId = 11U,
				.fov = 200U,
#else
				.cameraId = 7U,
				.fov = 200U,
#endif
				.fps = 30,
				.width = 1920U,
				.height = 1080U,
				.streamDepth = 3U,
				.clusterId = 3U,
		}
};


mdc::visual::Image GetImage(uint32_t camera_id)
{
	mdc::visual::Image image;
	// std::string base_prefix = "mdc_camera_instance_";
	image.header.frameId = "mdc_camera_instance_71";
	//        image.header.frameId = base_prefix + std::to_string(camera_id);
	image.frameData = imageData;
	return image;
}

#ifdef CONF_SOC_A
//HuCam cameraSample(camera_configs, 11);
#else
// HuCam cameraSample(camera_configs, 4);
#endif
} /*end of namespace*/


Image::Image ()
{
    using namespace ara::com;
    using namespace ara::core;
    using Skeleton = huawei::ap::sample::skeleton::VisualizationSkeleton;

    // 注意创建Skeleton的第1个参数是InstanceSpecifier类型，其值为配置文件network_binding.json中instanceSpecifier属性的值
    skeleton_ = std::make_shared<Skeleton>(InstanceSpecifier(
        "camera_h265_image_server/camera_h265_image_server/CameraH265ImagePPort"),
        MethodCallProcessingMode::kEvent);
}

mdc::visual::Image Image::getCameraImage(uint8_t *buffer, size_t length, uint32_t camera_id)
{
	mdc::visual::Image image;
	image.header.frameId = "mdc_camera_instance_71";
	image.frameData.assign(buffer, buffer+length);
	return image;
}

bool Image::inputParamsValid()
{
	if(live_cam_id == -1)
	{
		return false;
	}
	else if(pub_topic.empty())
	{
		return false;
	}
	else if (channel_number == 255)
	{
		return false;
	}
	else
	{
		return true;
	}
}

void Image::setInputParams(int liveCamId, std::string publishTopic, hi_venc_chn EncChannelNumber, std::string sampleName_,bool aclInit)
{
	live_cam_id = liveCamId;
	pub_topic = publishTopic;
	channel_number = EncChannelNumber;
	acl_init = aclInit;
}

void Image::CameraWorkingThread(int camera_id)
{
	while (!isStop) {
		CamByIndex(camera_id);
		// FlushTest(camera_id);
	}
}

void Image::MainWorkingThread()
{
	while (!isStop) {
        OnSend();
    }
}

void Image::Init()
{
	// 发布服务
	//VizSample::Init();
    skeleton_->OfferService();

	printf("here %d\n\n\n", __LINE__);
	mainThread = std::thread(&Image::MainWorkingThread, this);
	for(uint8_t i = 0; i<4; i++)
	{
		// cameraThread[i] = std::thread(&Image::FlushTest, this, i);
		if(availableCameras[i])
		{
			cameraThread[i] = std::thread(&Image::CameraWorkingThread, this, i);
			printf("[INFO] Thread for camera %d created \n", i);
		}
	}
}


//void Image::setupSenders()
//{
//	for (i = 0; i < 4; i++) {
//		sampleCameraImageForBanner[i] = skeleton_->CameraImageAsyncEvent.Allocate();
//	}
//}

void Image::setupPublishers()
{
	for(int i = 0; i<4; i++)
	{
		std::string topic = topic_prefix + std::to_string(i+cam_idx_div_factor);
		imagePub[i]  = mdc::visual::Publisher::Advertise<mdc::visual::Image>(topic);
	}
}

// uint8_t Image::getRawCameraId(uint8_t id) 
// {

// }

void Image::Publish(uint8_t camera_id, mdc::visual::Image image)
{
	timespec ts;
	clock_gettime(CLOCK_VIRTUAL, &ts);
	auto sampleCameraImageForBanner = skeleton_->CameraH265ImagePDataViz.Allocate();
	sampleCameraImageForBanner->cameraId = camera_id;
	sampleCameraImageForBanner->dataSize = 20;
	sampleCameraImageForBanner->header.stampHigh = ts.tv_sec;
	sampleCameraImageForBanner->header.stampLow = ts.tv_nsec;
	sampleCameraImageForBanner->header.frameId = image.header.frameId;
	sampleCameraImageForBanner->rawData= image.frameData;

	// 发送Event数据包
	skeleton_->CameraH265ImagePDataViz.Send(std::move(sampleCameraImageForBanner));
}

void Image::RunThread()
{
	static int a = 0;
	a++;
}

void Image::showBanner()
{
	printf("here %d\n\n\n", __LINE__);

	int i = 0;

	while(i++<bannerWaitStep)
	{
		for(int j = 0; j<4; j++)
		{

			mdc::visual::Image myImage = GetImage(10);
			Publish(j, myImage);
			// auto sampleCameraImageForBanner = skeleton_->CameraImageAsyncEvent.Allocate();
			// printf("here %d\n\n\n", __LINE__);

			// images[j] = GetImage(10);
			// sampleCameraImageForBanner->cameraId = j;
			// sampleCameraImageForBanner->dateSize = 20;
			// sampleCameraImageForBanner->header.stampSec = SystemUtils::GetCurrentTime().sec;
			// sampleCameraImageForBanner->header.stampNsec = SystemUtils::GetCurrentTime().nsec;
			// sampleCameraImageForBanner->header.frameId = images[j].header.frameId;
			// sampleCameraImageForBanner->frameData= images[j].frameData;
			// printf("here %d\n\n\n", __LINE__);

			// // 发送Event数据包
			// skeleton_->CameraImageAsyncEvent.Send(std::move(sampleCameraImageForBanner));
			// printf("here %d\n\n\n", __LINE__);

			log_.LogInfo() << "CameraImageAsyncEvent send ehsan";
		}
	}
}

bool Image::OnInit()
{ 
	if(!inputParamsValid())
	{
		return false;
	}
	setupPublishers();

	int i = 0;

	imageData = FileUtils::ReadBinaryDataFile(SystemUtils::GetFilePathInDataDir("image_for_overlay.h265"));
	if (imageData.empty()) {
		printf("[ERROR] no image found\n");
		return false;
	}
	showBanner();

	return true;
}


void Image::setHucamInstance(HuCam *cameraInstance)
{
	cameraSample = cameraInstance;
}

HuCam* Image::getHucamObjectPtr()
{
	return cameraSample;
}

void Image::Stop()
{
	skeleton_->StopOfferService();
}

void Image::OnSend()
{
	auto currentTime = std::chrono::high_resolution_clock::now();

	static bool hucam_inited = false;


	if(cameraSample == nullptr)
	{
		cameraSample = new HuCam();
		cameraSample->InitAcl();
		cameraSample->Init(camera_configs, -1);

	}
	else
	{
		if(!hucam_inited){
			first_is_init = true;
			bool allInit = true;

			int k = 0;
			for(k = 0; k< 4; k++)
			{
				if(availableCameras[k]) {
					if((!thread_inited[k]))
					{
						allInit = false;
						break;
					}
				}
			}
			if(allInit)
			{
				hucam_inited = true;
			}
			else
			{
				printf("[INFO] Waiting for thread %d to be initted. \n", k);
			}
		}
		else
		{
			cameraSample->getAllCamData();
		}
	}

	auto endTime = std::chrono::high_resolution_clock::now();
	auto duration = std::chrono::duration_cast<std::chrono::milliseconds>(endTime - currentTime);
	const int delaySleep = 33 - int(duration.count());

	std::this_thread::sleep_for(std::chrono::milliseconds(delaySleep));

}

void Image::FlushTest(int camera_id)
{
	images[camera_id] = GetImage(10);
	images[camera_id].header.stamp = SystemUtils::GetCurrentTime();
	imagePub[camera_id].Publish(images[camera_id]);
	std::this_thread::sleep_for(std::chrono::milliseconds(30));
	printf("[INFO] publish test image for camera %d. \n", camera_id);
}

void Image::CamByIndex(int camera_id)
{
	// int camera_id = 0;
	auto currentTime = std::chrono::high_resolution_clock::now();
	void* buff = nullptr;
	uint8_t *output_buff;
	size_t output_size;

	std::shared_ptr<hi_venc_stream> output_enc;


	static bool hucam_inited = false;

	if(first_is_init){
		if(!hucam_inited)
		{
			if(cameraSample != nullptr){
				cameraSample->InitAcl();
				cameraSample->Init(camera_configs, camera_id);
				cameraSample->createChannel(width, height, bufferSize, camera_id);
				hucam_inited = true;
				thread_inited[camera_id] = true;
			}
			else
			{
				printf("[WARNING] Waiting for camera sample to be created on first thread \n");
			}
		}
		else
		{
			if(!cameraSample->getEncoddedImageByIndex(&output_enc, camera_id))
			{
				if(1/*camera_id == 1*/){
				mdc::visual::Image myImage;
				myImage = getCameraImage((uint8_t *)output_enc->pack[0].addr, (size_t)output_enc->pack[0].len, CAMERA_5);
				Publish(camera_id, myImage);	
				}
				
				// images[camera_id] = getCameraImage((uint8_t *)output_enc->pack[0].addr, (size_t)output_enc->pack[0].len, CAMERA_5);
				// images[camera_id].header.stamp = SystemUtils::GetCurrentTime();
				// imagePub[camera_id].Publish(images[camera_id]);
				printf("[INFO] published image %d with size = %d \n", camera_id,images[camera_id].frameData.size());
				printf("in Publish dev factor is %d \n", cam_idx_div_factor);
			}
			else
			{
				printf("[WARNING] couldn't get encodded image by index \n");
			}
		}
	}
	else
	{
		printf("[WARNING] Waiting for first thread \n");
	}
	auto endTime = std::chrono::high_resolution_clock::now();
	auto duration = std::chrono::duration_cast<std::chrono::milliseconds>(endTime - currentTime);
	//printf("[INFO] Encoding Speed for camera %d is = %dms \n", camera_id, duration);
	int delaySleep = camSleepValueDefault - int(duration.count());
	delaySleep = (delaySleep > 0) ? delaySleep : 0;


	std::this_thread::sleep_for(std::chrono::milliseconds(delaySleep));
}



//Image* getImageByName(const std::string& targetSampleName) {
//    const std::vector<VizSamplePtr>& samples = SampleContainer::Instance().GetSamples();
//
//    auto it = std::find_if(samples.begin(), samples.end(), [&targetSampleName](const VizSamplePtr& sample) {
//        // Assuming Image is derived from VizSample and has a member named sampleName
//        Image* image = dynamic_cast<Image*>(sample.get());
//        return image != nullptr && image->getSampleName() == targetSampleName;
//    });
//
//    if (it != samples.end()) {
//        // Found an Image with the specified name
//        return dynamic_cast<Image*>(it->get());
//    }
//
//    // Return nullptr if not found
//    return nullptr;
//}


//std::unique_ptr<Image> img4 = std::make_unique<Image>();
//
//
//static VizSampleRegister g_imageReg4([]() {
//    img4->setInputParams(8, "mdc_camera_instance_4", 0, "cam4Sample" ,true);
//    return std::move(img4);

// });


